"""
copy from:
https://github.com/shenweichen/DeepCTR-Torch/blob/b4d8181e86c2165722fa9331bc16185832596232/deepctr_torch/callbacks.py
"""
import torch
from tensorflow.python.keras.callbacks import EarlyStopping
from tensorflow.python.keras.callbacks import ModelCheckpoint
from tensorflow.python.keras.callbacks import History

EarlyStopping = EarlyStopping
History = History

class ModelCheckpoint(ModelCheckpoint):
    """Save the model after every epoch.
    `filepath` can contain named formatting options,
    which will be filled the value of `epoch` and
    keys in `logs` (passed in `on_epoch_end`).
    For example: if `filepath` is `weights.{epoch:02d}-{val_loss:.2f}.hdf5`,
    then the model checkpoints will be saved with the epoch number and
    the validation loss in the filename.
    Arguments:
        filepath: string, path to save the model file.
        monitor: quantity to monitor.
        verbose: verbosity mode, 0 or 1.
        save_best_only: if `save_best_only=True`,
            the latest best model according to
            the quantity monitored will not be overwritten.
        mode: one of {auto, min, max}.
            If `save_best_only=True`, the decision
            to overwrite the current save file is made
            based on either the maximization or the
            minimization of the monitored quantity. For `val_acc`,
            this should be `max`, for `val_loss` this should
            be `min`, etc. In `auto` mode, the direction is
            automatically inferred from the name of the monitored quantity.
        save_weights_only: if True, then only the model's weights will be
            saved (`model.save_weights(filepath)`), else the full model
            is saved (`model.save(filepath)`).
        period: Interval (number of epochs) between checkpoints.
    """

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0
            filepath = self.filepath.format(epoch=epoch + 1, **logs)
            if self.save_best_only:
                current = logs.get(self.monitor)
                if current is None:
                    print('Can save best model only with %s available, skipping.' % self.monitor)
                else:
                    if self.monitor_op(current, self.best):
                        if self.verbose > 0:
                            print('Epoch %05d: %s improved from %0.5f to %0.5f,'
                                  ' saving model to %s' % (epoch + 1, self.monitor, self.best,
                                                           current, filepath))
                        self.best = current
                        if self.save_weights_only:
                            torch.save(self.model.state_dict(), filepath)
                        else:
                            torch.save(self.model, filepath)
                    else:
                        if self.verbose > 0:
                            print('Epoch %05d: %s did not improve from %0.5f' %
                                  (epoch + 1, self.monitor, self.best))
            else:
                if self.verbose > 0:
                    print('Epoch %05d: saving model to %s' %
                          (epoch + 1, filepath))
                if self.save_weights_only:
                    torch.save(self.model.state_dict(), filepath)
                else:
                    torch.save(self.model, filepath)


from torch.utils.data import Dataset, DataLoader
from deepctr_torch.inputs import SparseFeat, VarLenSparseFeat
import torch.nn as nn


class CommonDataSet(Dataset):
    # TODO, 可以生成完整的 训练数据生成逻辑
    def __init__(self) -> None:
        super().__init__()
        
        self.data = None
    
    def __getitem__(self, index):
        return self.data[index]
        
    def __len__(self):
        return len(self.data)

# copy from DeepMatch and remove device
def create_embedding_matrix(feature_columns, init_std=0.001, linear=False, sparse=False, device='cpu'):
    # Return nn.ModuleDict: for sparse features, {embedding_name: nn.Embedding}
    # for varlen sparse features, {embedding_name: nn.EmbeddingBag}
    sparse_feature_columns = list(
        filter(lambda x: isinstance(x, SparseFeat), feature_columns)) if len(feature_columns) else []

    # for feat in sparse_feature_columns:
    #     print(feat.embedding_name)
    varlen_sparse_feature_columns = list(
        filter(lambda x: isinstance(x, VarLenSparseFeat), feature_columns)) if len(feature_columns) else []

    embedding_dict = nn.ModuleDict(
        {feat.embedding_name: nn.Embedding(feat.vocabulary_size, feat.embedding_dim if not linear else 1, sparse=sparse)
         for feat in
         sparse_feature_columns + varlen_sparse_feature_columns}
    )

    # for feat in varlen_sparse_feature_columns:
    #     embedding_dict[feat.embedding_name] = nn.EmbeddingBag(
    #         feat.dimension, embedding_size, sparse=sparse, mode=feat.combiner)

    for tensor in embedding_dict.values():
        nn.init.normal_(tensor.weight, mean=0, std=init_std)

    return embedding_dict




import torch.nn.functional as F

from .PLBaseModel import PLBaseModel
from deepctr_torch.layers import DNN
from ..utils import combined_dnn_input


class DSSM(PLBaseModel):
    def __init__(self, user_feature_columns, item_feature_columns, 
                dnn_hidden_units, dnn_activation="relu", gamma=0.01,
                l2_reg_linear=0.00001, l2_reg_embedding=0.00001, init_std=0.0001, seed=1024, task='binary', device='cpu', **kwargs):
        super().__init__(user_feature_columns, item_feature_columns, l2_reg_linear=l2_reg_linear, l2_reg_embedding=l2_reg_embedding, init_std=init_std, seed=seed, task=task, device=device, **kwargs)

        self.user_dnn = DNN(self.compute_input_dim(user_feature_columns), dnn_hidden_units,
                        activation=dnn_activation, init_std=init_std, device=device)

        self.item_dnn = DNN(self.compute_input_dim(item_feature_columns), dnn_hidden_units,
                            activation=dnn_activation, init_std=init_std, device=device)

    def forward(self, inputs):
        item_embedding = self.item_tower(inputs)
        user_embedding = self.user_tower(inputs)
        if self.mode == "user_representation":
            return user_embedding
        if self.mode == "item_representation":
            return item_embedding

        score = F.cosine_similarity(item_embedding, user_embedding, dim=1)
        # shape is (batch)
        return score
    
    def user_tower(self, inputs):
        if self.mode == "item_representation":
            return None
        user_sparse_embedding_list, user_dense_value_list = \
            self.input_from_feature_columns(inputs, self.user_feature_columns, self.embedding_dict)

        user_dnn_input = combined_dnn_input(user_sparse_embedding_list, user_dense_value_list)
        user_embedding = self.user_dnn(user_dnn_input)
        return user_embedding

    def item_tower(self, inputs):
        if self.mode == "user_representation":
            return None
        item_sparse_embedding_list, item_dense_value_list = \
            self.input_from_feature_columns(inputs, self.item_feature_columns, self.embedding_dict)

        item_dnn_input = combined_dnn_input(item_sparse_embedding_list, item_dense_value_list)
        item_embedding = self.item_dnn(item_dnn_input)
        return item_embedding


        


from abc import abstractmethod
from functools import partial

import torch.nn.functional as F
from pytorch_lightning import LightningModule, Trainer
from torch.utils.data import DataLoader, TensorDataset

import numpy as np
import torch
import torch.nn as nn
from deepctr_torch.inputs import build_input_features
from deepctr_torch.inputs import varlen_embedding_lookup, get_varlen_pooling_list
from deepctr_torch.inputs import SparseFeat, DenseFeat, VarLenSparseFeat
from ..inputs import create_embedding_matrix


# DeepMatch_Torch Linear Model
class Linear(nn.Module):
    def __init__(self, feature_columns, feature_index, init_std=0.0001, device='cpu'):
        super(Linear, self).__init__()
        self.my_device = device
        self.feature_index = feature_index
        self.sparse_feature_columns = list(
            filter(lambda x: isinstance(x, SparseFeat), feature_columns)) if len(feature_columns) else []
        self.dense_feature_columns = list(
            filter(lambda x: isinstance(x, DenseFeat), feature_columns)) if len(feature_columns) else []

        self.varlen_sparse_feature_columns = list(
            filter(lambda x: isinstance(x, VarLenSparseFeat), feature_columns)) if len(feature_columns) else []

        self.embedding_dict = create_embedding_matrix(feature_columns, init_std, linear=True, sparse=False,)
        for tensor in self.embedding_dict.values():
            nn.init.normal_(tensor.weight, mean=0, std=init_std)

        if len(self.dense_feature_columns) > 0:
            self.weight = nn.Parameter(torch.Tensor(sum(fc.dimension for fc in self.dense_feature_columns), 1))
            torch.nn.init.normal_(self.weight, mean=0, std=init_std)

    def forward(self, X, sparse_feat_refine_weight=None):

        sparse_embedding_list = [self.embedding_dict[feat.embedding_name](
            X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]].long()) for
            feat in self.sparse_feature_columns]

        dense_value_list = [X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]] for feat in
                            self.dense_feature_columns]

        sequence_embed_dict = varlen_embedding_lookup(X, self.embedding_dict, self.feature_index,
                                                      self.varlen_sparse_feature_columns)
        varlen_embedding_list = get_varlen_pooling_list(sequence_embed_dict, X, self.feature_index,
                                                        self.varlen_sparse_feature_columns, self.my_device)

        sparse_embedding_list += varlen_embedding_list

        linear_logit = torch.zeros([X.shape[0], 1]).to(sparse_embedding_list[0].device)
        if len(sparse_embedding_list) > 0:
            sparse_embedding_cat = torch.cat(sparse_embedding_list, dim=-1)
            if sparse_feat_refine_weight is not None:
                sparse_embedding_cat = sparse_embedding_cat * sparse_feat_refine_weight.unsqueeze(1)
            sparse_feat_logit = torch.sum(sparse_embedding_cat, dim=-1, keepdim=False)
            linear_logit += sparse_feat_logit
        if len(dense_value_list) > 0:
            dense_value_logit = torch.cat(
                dense_value_list, dim=-1).matmul(self.weight)
            linear_logit += dense_value_logit

        return linear_logit


# 使用 Pl 作为 fit 的接口
class PLBaseModel(LightningModule):
    """Base class for all DeepMatch_Torch models.
    This model inspired from: https://github.com/Rose-STL-Lab/torchTS/blob/main/torchts/nn/model.py
    Args:
        optimizer (torch.optim.Optimizer): Optimizer
        opimizer_args (dict): Arguments for the optimizer
        criterion: Loss function
        criterion_args (dict): Arguments for the loss function
        scheduler (torch.optim.lr_scheduler): Learning rate scheduler
        scheduler_args (dict): Arguments for the scheduler
        scaler (torchts.utils.scaler.Scaler): Scaler
    """

    def __init__(
        self,
        user_feature_columns,
        item_feature_columns,
        optimizer=None,
        optimizer_args=None,
        criterion=F.mse_loss,
        criterion_args=None,
        scheduler=None,
        scheduler_args=None,
        scaler=None,
        config={},
        **kwargs,
    ):
        super().__init__()
        self.user_feature_columns = user_feature_columns
        self.item_feature_columns = item_feature_columns
        # DeepMatch side 用于增加兼容
        self.config = config  # 增加 config 用于设置 DeepMatch 侧需要的参数
        # 优先使用 kwargs 的配置
        # TODO: 以后要消灭所有的 device , 
        self.config.update(kwargs)
        
        # 用于判断模型 输出 logits 还是 user/item 的 vector 表示
        self.mode = self.config.get('mode', 'train')

        self.linear_feature_columns = user_feature_columns + item_feature_columns
        self.dnn_feature_columns = self.linear_feature_columns 

        # 在 pl 中不需要 to(device)
        self.reg_loss = torch.zeros((1,))
        self.aux_loss = torch.zeros((1,))

        self.feature_index = build_input_features(
            self.linear_feature_columns)

        self.embedding_dict = create_embedding_matrix(self.dnn_feature_columns, 
                    self.config.get("init_std"), sparse=False, )

        self.linear_model = Linear(
            self.linear_feature_columns, self.feature_index, device=self.device)

        self.regularization_weight = []

        self.add_regularization_weight(self.embedding_dict.parameters(), 
            l2=self.config.get("l2_reg_embedding"))
        self.add_regularization_weight(self.linear_model.parameters(), 
            l2=self.config.get("l2_reg_linear"))

        self.criterion = criterion
        self.criterion_args = criterion_args
        self.scaler = scaler
        self.gpus = self.config.get("gpus", 0)

        # 增加了选择 optimizer 的方式
        optimizer = self.init_optimizer(optimizer)
        if optimizer_args is not None:
            self.optimizer = partial(optimizer, **optimizer_args)
        else:
            self.optimizer = optimizer

        if scheduler is not None and scheduler_args is not None:
            self.scheduler = partial(scheduler, **scheduler_args)
        else:
            self.scheduler = scheduler

    def fit(self, x, y, max_epochs=10, batch_size=128):
        """Fits model to the given data.
        Args:
            x (torch.Tensor): Input data
            y (torch.Tensor): Output data
            max_epochs (int): Number of training epochs
            batch_size (int): Batch size for torch.utils.data.DataLoader
        """
        x, y = self.construct_input(x, y)
        dataset = TensorDataset(x, y)
        loader = DataLoader(dataset, batch_size=batch_size, shuffle=True)
        trainer = Trainer(max_epochs=max_epochs, gpus=self.gpus)
        trainer.fit(self, loader)

    def prepare_batch(self, batch):
        return batch

    def _step(self, batch, batch_idx, num_batches):
        """
        Args:
            batch: Output of the torch.utils.data.DataLoader
            batch_idx: Integer displaying index of this batch
            dataset: Data set to use
        Returns: loss for the batch
        """
        x, y = self.prepare_batch(batch)

        if self.training:
            batches_seen = batch_idx + self.current_epoch * num_batches
        else:
            batches_seen = batch_idx

        pred = self(x)

        if self.scaler is not None:
            y = self.scaler.inverse_transform(y)
            pred = self.scaler.inverse_transform(pred)

        # 假设是 softmax, 应该把 y 设置为 Long类型
        if self.criterion == F.cross_entropy:
            y = y.long()
        if self.criterion_args is not None:
            loss = self.criterion(pred, y, **self.criterion_args)
        else:
            loss = self.criterion(pred, y)

        return loss

    def training_step(self, batch, batch_idx):
        """Trains model for one step.
        Args:
            batch (torch.Tensor): Output of the torch.utils.data.DataLoader
            batch_idx (int): Integer displaying index of this batch
        """
        train_loss = self._step(batch, batch_idx, len(self.trainer.train_dataloader))
        self.log(
            "train_loss",
            train_loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        return train_loss

    def validation_step(self, batch, batch_idx):
        """Validates model for one step.
        Args:
            batch (torch.Tensor): Output of the torch.utils.data.DataLoader
            batch_idx (int): Integer displaying index of this batch
        """
        val_loss = self._step(batch, batch_idx, len(self.trainer.val_dataloader))
        self.log("val_loss", val_loss)
        return val_loss

    def test_step(self, batch, batch_idx):
        """Tests model for one step.
        Args:
            batch (torch.Tensor): Output of the torch.utils.data.DataLoader
            batch_idx (int): Integer displaying index of this batch
        """
        test_loss = self._step(batch, batch_idx, len(self.trainer.test_dataloader))
        self.log("test_loss", test_loss)
        return test_loss

    @abstractmethod
    def forward(self, x):
        """Forward pass.
        Args:
            x (torch.Tensor): Input data
        Returns:
            torch.Tensor: Predicted data
        """

    def predict(self, x):
        """Runs model inference.
        Args:
            x (torch.Tensor): Input data
        Returns:
            torch.Tensor: Predicted data
        """
        return self(x).detach()

    def configure_optimizers(self):
        """Configure optimizer.
        Returns:
            torch.optim.Optimizer: Optimizer
        """
        optimizer = self.optimizer(self.parameters())

        if self.scheduler is not None:
            scheduler = self.scheduler(optimizer)
            return [optimizer], [scheduler]

        return optimizer

    def init_optimizer(self, optimizer_or_name):
        """init_optimizer 假设 optimizer 是一个string, 那么返回 torch.optim对应的 optimizer
            否则，返回自己
        Args:
            optimizer_or_name ([type]): [description]

        Returns:
            [type]: [description]
        """
        if isinstance(optimizer_or_name, str):
            if optimizer_or_name == "Adam":
                return torch.optim.Adam
            elif optimizer_or_name == "SGD":
                return torch.optim.SGD
        else:
            return optimizer_or_name
            
    def construct_input(self, x, y):
        # 有可能直接输出的就是 tensor
        if isinstance(x, dict):
            x = [x[feature] for feature in self.feature_index]
        
        for i in range(len(x)):
            if len(x[i].shape) == 1:
                x[i] = np.expand_dims(x[i], axis=1)
        x = torch.from_numpy(np.concatenate(x, axis=-1))
        y = torch.tensor(y, dtype=torch.float32)
        return x, y

    def add_regularization_weight(self, weight_list, l1=0.0, l2=0.0):
        # For a Parameter, put it in a list to keep Compatible with get_regularization_loss()
        if isinstance(weight_list, torch.nn.parameter.Parameter):
            weight_list = [weight_list]
        # For generators, filters and ParameterLists, convert them to a list of tensors to avoid bugs.
        # e.g., we can't pickle generator objects when we save the model.
        else:
            weight_list = list(weight_list)
        self.regularization_weight.append((weight_list, l1, l2))

    def get_regularization_loss(self, ):
        total_reg_loss = torch.zeros((1,), )
        for weight_list, l1, l2 in self.regularization_weight:
            for w in weight_list:
                if isinstance(w, tuple):
                    parameter = w[1]  # named_parameters
                else:
                    parameter = w
                if l1 > 0:
                    total_reg_loss += torch.sum(l1 * torch.abs(parameter))
                if l2 > 0:
                    try:
                        total_reg_loss += torch.sum(l2 * torch.square(parameter))
                    except AttributeError:
                        total_reg_loss += torch.sum(l2 * parameter * parameter)

        return total_reg_loss
    
    def input_from_feature_columns(self, X, feature_columns, embedding_dict, support_dense=True):
        
        sparse_feature_columns = list(
            filter(lambda x: isinstance(x, SparseFeat), feature_columns)) if len(feature_columns) else []
        dense_feature_columns = list(
            filter(lambda x: isinstance(x, DenseFeat), feature_columns)) if len(feature_columns) else []

        varlen_sparse_feature_columns = list(
            filter(lambda x: isinstance(x, VarLenSparseFeat), feature_columns)) if feature_columns else []

        if not support_dense and len(dense_feature_columns) > 0:
            raise ValueError(
                "DenseFeat is not supported in dnn_feature_columns")

        sparse_embedding_list = [embedding_dict[feat.embedding_name](
            X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]].long()) for
            feat in sparse_feature_columns]
        sequence_embed_dict = varlen_embedding_lookup(X, self.embedding_dict, self.feature_index,
                                                      varlen_sparse_feature_columns)
        varlen_sparse_embedding_list = get_varlen_pooling_list(sequence_embed_dict, X, self.feature_index,
                                                               varlen_sparse_feature_columns, self.device)
        dense_value_list = [X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]] for feat in
                            dense_feature_columns]

        return sparse_embedding_list + varlen_sparse_embedding_list, dense_value_list

    def compute_input_dim(self, feature_columns, include_sparse=True, include_dense=True, feature_group=False):
        sparse_feature_columns = list(
            filter(lambda x: isinstance(x, (SparseFeat, VarLenSparseFeat)), feature_columns)) if len(
            feature_columns) else []
        dense_feature_columns = list(
            filter(lambda x: isinstance(x, DenseFeat), feature_columns)) if len(feature_columns) else []

        dense_input_dim = sum(
            map(lambda x: x.dimension, dense_feature_columns))
        if feature_group:
            sparse_input_dim = len(sparse_feature_columns)
        else:
            sparse_input_dim = sum(feat.embedding_dim for feat in sparse_feature_columns)
        input_dim = 0
        if include_sparse:
            input_dim += sparse_input_dim
        if include_dense:
            input_dim += dense_input_dim
        return input_dim

    def rebuild_feature_index(self, feature_columns):
        # 为了满足 单独预测 user/item vector 的需求，需要重新知道 feature_columns 的位置
        self.feature_index = build_input_features(feature_columns)
        return self

    def full_predict(self, x, batch_size=256):
        # 有可能直接输入的就是 tensor
        if isinstance(x, dict):
            x = [x[feature] for feature in self.feature_index]
        
        for i in range(len(x)):
            if len(x[i].shape) == 1:
                x[i] = np.expand_dims(x[i], axis=1)
        x = torch.from_numpy(np.concatenate(x, axis=-1))
        dataset = TensorDataset(x)
        loader = DataLoader(dataset, batch_size=batch_size, shuffle=True)
        ret = []
        for batch in loader:            
            tmp_result = self.predict(batch[0])
            ret.append(tmp_result)
        return torch.cat(ret, axis=0)
        



from .PLBaseModel import PLBaseModel
from deepctr_torch.layers import DNN
from deepctr_torch.inputs import SparseFeat, DenseFeat, VarLenSparseFeat
from deepctr_torch.inputs import get_varlen_pooling_list, varlen_embedding_lookup
from ..utils import combined_dnn_input
from ..inputs import create_embedding_matrix
import torch.nn.functional as F


# youtube dnn 的结构没有什么改变
class YouTubeDNN(PLBaseModel):
    def __init__(self, user_feature_columns, item_feature_columns, 
                num_sampled=5, user_dnn_hidden_units=[64, 32], 
                dnn_activation='relu', dnn_use_bn=False,
                device="cpu", init_std=0.002,
               l2_reg_dnn=0, l2_reg_embedding=1e-6, 
               dnn_dropout=0, activation='relu', seed=1024, **kwargs):
        super(YouTubeDNN, self).__init__(user_feature_columns, item_feature_columns, 
                l2_reg_linear=1e-5, l2_reg_embedding=1e-5,
                init_std=0.0001, seed=1024, task='binary', device='cpu', **kwargs)
        self.num_sampled = num_sampled

        self.user_dnn = DNN(self.compute_input_dim(user_feature_columns), user_dnn_hidden_units,
                        activation=dnn_activation, init_std=init_std, device=device)

    def forward(self, X):
        batch_size = X.size(0)
        user_embedding = self.user_tower(X)
        item_embedding = self.item_tower(X)

        if self.mode == "user_representation":
            return user_embedding
        if self.mode == "item_representation":
            return item_embedding

        score = F.cosine_similarity(user_embedding, item_embedding, dim=-1)
        score = score.view(batch_size, -1)
        return score
    
    def item_tower(self, X):
        if self.mode == "user_representation":
            return None
        
        item_embedding_list, _ = self.input_from_item_feature_columns(X, self.item_feature_columns, self.embedding_dict)
        item_embedding = item_embedding_list[0]  # (batch, movie_list_len, feat_dim)
        return item_embedding

    def user_tower(self, X):
        if self.mode == "item_representation":
            return None
        # sample softmax 可以通过 构造样本实现
        user_sparse_embedding_list, user_dense_value_list = \
            self.input_from_feature_columns(X, self.user_feature_columns, self.embedding_dict)
        
        user_dnn_input = combined_dnn_input(user_sparse_embedding_list, user_dense_value_list)
        user_embedding = self.user_dnn(user_dnn_input)  # (batch_size, embedding_dim)
        user_embedding = user_embedding.unsqueeze(1) # (batch, 1, embedding_dim)
        return user_embedding
    
    def input_from_item_feature_columns(self, X, feature_columns, embedding_dict, support_dense=True):
    
        sparse_feature_columns = list(
            filter(lambda x: isinstance(x, SparseFeat), feature_columns)) if len(feature_columns) else []
        dense_feature_columns = list(
            filter(lambda x: isinstance(x, DenseFeat), feature_columns)) if len(feature_columns) else []

        varlen_sparse_feature_columns = list(
            filter(lambda x: isinstance(x, VarLenSparseFeat), feature_columns)) if feature_columns else []

        if not support_dense and len(dense_feature_columns) > 0:
            raise ValueError(
                "DenseFeat is not supported in dnn_feature_columns")

        sparse_embedding_list = [embedding_dict[feat.embedding_name](
            X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]].long()) for
            feat in sparse_feature_columns]

        # 这里返回的就是 movie_id 的 embedding
        sequence_embed_dict = varlen_embedding_lookup(X, self.embedding_dict, self.feature_index,
                                                      varlen_sparse_feature_columns)
        feat_name = varlen_sparse_feature_columns[0].name
        item_embedding = sequence_embed_dict[feat_name] 
        # shape is (batch, movie_id_len, feat_dim)

        varlen_sparse_embedding_list = [item_embedding]
        dense_value_list = [X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]] for feat in
                            dense_feature_columns]

        return sparse_embedding_list + varlen_sparse_embedding_list, dense_value_list


import torch

from .PLBaseModel import PLBaseModel
from ..layers import FactorizationMachines


class FM(PLBaseModel):
    def __init__(self, user_feature_columns, item_feature_columns, 
        l2_reg_linear=0.00001, l2_reg_embedding=0.00001, init_std=0.0001, 
        seed=1024, task='binary', device='cpu', gpus=None, **kwargs):
        super().__init__(user_feature_columns, item_feature_columns, l2_reg_linear=l2_reg_linear, l2_reg_embedding=l2_reg_embedding, init_std=init_std, seed=seed, task=task, device=device, gpus=gpus, **kwargs)
        self.fm = FactorizationMachines()

    def forward(self, X):
        # (batch, N) N 表示 多列
        # X 包括两部分，一部分是 linear part, 一部分是 fm
        if self.mode == 'train':
            linear_logit = self.linear_model(X)
        #  TODO: 增加 dense_value interaction
        user_fm_input = self.user_tower(X)
        item_fm_input = self.item_tower(X)
        if self.mode == "user_representation":
            return user_fm_input
        if self.mode == "item_representation":
            return item_fm_input
        
        fm_input = torch.cat([user_fm_input, item_fm_input], dim=1)
        # 需要确定的是那些 fields 需要 intersection
        fm_logit = self.fm(fm_input)

        res = linear_logit + fm_logit
        # shape is (batch_size, 1) predict 部分进行了 squeeze 
        res = res.squeeze()
        return res
    
    def item_tower(self, X):
        if self.mode == "user_representation":
            return None
        sparse_embedding_list, dense_value_list = self.input_from_feature_columns(X, self.item_feature_columns,
                                                                                  self.embedding_dict)
        fm_input = torch.cat(sparse_embedding_list, dim=1)
        return fm_input

    def user_tower(self, X):
        if self.mode == "item_representation":
            return None
        sparse_embedding_list, dense_value_list = self.input_from_feature_columns(X, self.user_feature_columns,
                                                                                  self.embedding_dict)
        fm_input = torch.cat(sparse_embedding_list, dim=1)
        return fm_input


from turtle import forward
from numpy import AxisError
import torch
import torch.nn as nn


class CapsuleLayer(nn.Module):
    """
    revised from: https://github.com/Riroaki/CapsNet/blob/master/capsnet.py
    """

    def __init__(self, in_dim, in_caps, out_caps, out_dim, num_routing):
        """
        Initialize the layer.
        Args:
            in_dim: 		Dimensionality of each capsule vector.
            in_caps: 		Number of input capsules if digits layer.
            out_caps: 		Number of capsules in the capsule layer
            out_dim: 		Dimensionality, of the output capsule vector.
            num_routing:	Number of iterations during routing algorithm
        """
        super(CapsuleLayer, self).__init__()
        self.in_dim = in_dim
        self.in_caps = in_caps
        self.out_caps = out_caps
        self.out_dim = out_dim
        self.num_routing = num_routing
        self.device = "cpu"
        self.W = nn.Parameter(0.01 * torch.randn(1, out_caps, in_caps, out_dim, in_dim),
                              requires_grad=True)

    def forward(self, x):
        batch_size = x.size(0)
        # (batch_size, in_caps, in_dim) -> (batch_size, 1, in_caps, in_dim, 1)
        x = x.unsqueeze(1).unsqueeze(4)
        # W @ x =
        # (1, out_caps, in_caps, out_dim, in_dim) @ (batch_size, 1, in_caps, in_dim, 1) =
        # (batch_size, out_caps, in_caps, out_dims, 1)
        u_hat = torch.matmul(self.W, x)
        # (batch_size, out_caps, in_caps, out_dim)
        u_hat = u_hat.squeeze(-1)
        # detach u_hat during routing iterations to prevent gradients from flowing
        temp_u_hat = u_hat.detach()

        b = torch.zeros(batch_size, self.out_caps, self.in_caps, 1).to(self.device)

        for route_iter in range(self.num_routing - 1):
            # (batch_size, out_caps, in_caps, 1) -> Softmax along out_caps
            c = b.softmax(dim=1)

            # element-wise multiplication
            # (batch_size, out_caps, in_caps, 1) * (batch_size, in_caps, out_caps, out_dim) ->
            # (batch_size, out_caps, in_caps, out_dim) sum across in_caps ->
            # (batch_size, out_caps, out_dim)
            s = (c * temp_u_hat).sum(dim=2)
            # apply "squashing" non-linearity along out_dim
            v = squash(s)
            # dot product agreement between the current output vj and the prediction uj|i
            # (batch_size, out_caps, in_caps, out_dim) @ (batch_size, out_caps, out_dim, 1)
            # -> (batch_size, out_caps, in_caps, 1)
            uv = torch.matmul(temp_u_hat, v.unsqueeze(-1))
            b += uv

        # last iteration is done on the original u_hat, without the routing weights update
        c = b.softmax(dim=1)
        s = (c * u_hat).sum(dim=2)
        # apply "squashing" non-linearity along out_dim
        v = squash(s)
        return v


def squash(x, dim=-1):
    squared_norm = (x ** 2).sum(dim=dim, keepdim=True)
    scale = squared_norm / (1 + squared_norm)
    return scale * x / (squared_norm.sqrt() + 1e-8)



def main():
    cp = CapsuleLayer(10, 1, 4, 5, 10)
    i =  torch.rand(4, 1, 10)

    cp(i)

main()



"""
copy from:
https://github.com/shenweichen/DeepCTR-Torch/blob/b4d8181e86c2165722fa9331bc16185832596232/deepctr_torch/models/basemodel.py
Author:
    Weichen Shen,weichenswc@163.com

changed By BBruceyuan
"""
from __future__ import print_function

import time

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data as Data
from sklearn.metrics import *
from torch.utils.data import DataLoader
from tqdm import tqdm
from deepctr_torch.inputs import build_input_features, create_embedding_matrix
from deepctr_torch.inputs import varlen_embedding_lookup, get_varlen_pooling_list
from deepctr_torch.inputs import SparseFeat, DenseFeat, VarLenSparseFeat
from deepctr_torch.layers import PredictionLayer
from deepctr_torch.layers.utils import slice_arrays
from ..callbacks import History

try:
    from tensorflow.python.keras.callbacks import CallbackList
except ImportError:
    from tensorflow.python.keras._impl.keras.callbacks import CallbackList


class Linear(nn.Module):
    def __init__(self, feature_columns, feature_index, init_std=0.0001, device='cpu'):
        super(Linear, self).__init__()
        self.feature_index = feature_index
        self.device = device
        self.sparse_feature_columns = list(
            filter(lambda x: isinstance(x, SparseFeat), feature_columns)) if len(feature_columns) else []
        self.dense_feature_columns = list(
            filter(lambda x: isinstance(x, DenseFeat), feature_columns)) if len(feature_columns) else []

        self.varlen_sparse_feature_columns = list(
            filter(lambda x: isinstance(x, VarLenSparseFeat), feature_columns)) if len(feature_columns) else []

        self.embedding_dict = create_embedding_matrix(feature_columns, init_std, linear=True, sparse=False,
                                                      device=device)

        for tensor in self.embedding_dict.values():
            nn.init.normal_(tensor.weight, mean=0, std=init_std)

        if len(self.dense_feature_columns) > 0:
            self.weight = nn.Parameter(torch.Tensor(sum(fc.dimension for fc in self.dense_feature_columns), 1).to(
                device))
            torch.nn.init.normal_(self.weight, mean=0, std=init_std)

    def forward(self, X, sparse_feat_refine_weight=None):

        sparse_embedding_list = [self.embedding_dict[feat.embedding_name](
            X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]].long()) for
            feat in self.sparse_feature_columns]

        dense_value_list = [X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]] for feat in
                            self.dense_feature_columns]

        sequence_embed_dict = varlen_embedding_lookup(X, self.embedding_dict, self.feature_index,
                                                      self.varlen_sparse_feature_columns)
        varlen_embedding_list = get_varlen_pooling_list(sequence_embed_dict, X, self.feature_index,
                                                        self.varlen_sparse_feature_columns, self.device)

        sparse_embedding_list += varlen_embedding_list

        linear_logit = torch.zeros([X.shape[0], 1]).to(sparse_embedding_list[0].device)
        if len(sparse_embedding_list) > 0:
            sparse_embedding_cat = torch.cat(sparse_embedding_list, dim=-1)
            if sparse_feat_refine_weight is not None:
                sparse_embedding_cat = sparse_embedding_cat * sparse_feat_refine_weight.unsqueeze(1)
            sparse_feat_logit = torch.sum(sparse_embedding_cat, dim=-1, keepdim=False)
            linear_logit += sparse_feat_logit
        if len(dense_value_list) > 0:
            dense_value_logit = torch.cat(
                dense_value_list, dim=-1).matmul(self.weight)
            linear_logit += dense_value_logit

        return linear_logit


class BaseModel(nn.Module):
    def __init__(self, 
                user_feature_columns, item_feature_columns, l2_reg_linear=1e-5, l2_reg_embedding=1e-5,
                init_std=0.0001, seed=1024, task='binary', device='cpu', gpus=None):

        super(BaseModel, self).__init__()
        
        torch.manual_seed(seed)
        self.user_feature_columns = user_feature_columns
        self.item_feature_columns = item_feature_columns
        
        self.linear_feature_columns = item_feature_columns + user_feature_columns
        self.dnn_feature_columns = self.linear_feature_columns 

        self.reg_loss = torch.zeros((1,), device=device)
        self.aux_loss = torch.zeros((1,), device=device)
        self.device = device
        self.gpus = gpus
        if gpus and str(self.gpus[0]) not in self.device:
            raise ValueError(
                "`gpus[0]` should be the same gpu with `device`")

        self.feature_index = build_input_features(
            self.linear_feature_columns)

        self.embedding_dict = create_embedding_matrix(self.dnn_feature_columns, init_std, sparse=False, device=device)

        self.linear_model = Linear(
            self.linear_feature_columns, self.feature_index, device=device)

        self.regularization_weight = []

        self.add_regularization_weight(self.embedding_dict.parameters(), l2=l2_reg_embedding)
        self.add_regularization_weight(self.linear_model.parameters(), l2=l2_reg_linear)

        self.out = PredictionLayer(task, )
        self.to(device)

        # parameters for callbacks
        self._is_graph_network = True  # used for ModelCheckpoint in tf2
        self._ckpt_saved_epoch = False  # used for EarlyStopping in tf1.14
        self.history = History()

    def fit(self, x=None, y=None, batch_size=None, epochs=1, verbose=1, initial_epoch=0, validation_split=0.,
            validation_data=None, shuffle=True, callbacks=None):
        """
        :param x: Numpy array of training data (if the model has a single input), or list of Numpy arrays (if the model has multiple inputs).If input layers in the model are named, you can also pass a
            dictionary mapping input names to Numpy arrays.
        :param y: Numpy array of target (label) data (if the model has a single output), or list of Numpy arrays (if the model has multiple outputs).
        :param batch_size: Integer or `None`. Number of samples per gradient update. If unspecified, `batch_size` will default to 256.
        :param epochs: Integer. Number of epochs to train the model. An epoch is an iteration over the entire `x` and `y` data provided. Note that in conjunction with `initial_epoch`, `epochs` is to be understood as "final epoch". The model is not trained for a number of iterations given by `epochs`, but merely until the epoch of index `epochs` is reached.
        :param verbose: Integer. 0, 1, or 2. Verbosity mode. 0 = silent, 1 = progress bar, 2 = one line per epoch.
        :param initial_epoch: Integer. Epoch at which to start training (useful for resuming a previous training run).
        :param validation_split: Float between 0 and 1. Fraction of the training data to be used as validation data. The model will set apart this fraction of the training data, will not train on it, and will evaluate the loss and any model metrics on this data at the end of each epoch. The validation data is selected from the last samples in the `x` and `y` data provided, before shuffling.
        :param validation_data: tuple `(x_val, y_val)` or tuple `(x_val, y_val, val_sample_weights)` on which to evaluate the loss and any model metrics at the end of each epoch. The model will not be trained on this data. `validation_data` will override `validation_split`.
        :param shuffle: Boolean. Whether to shuffle the order of the batches at the beginning of each epoch.
        :param callbacks: List of `deepctr_torch.callbacks.Callback` instances. List of callbacks to apply during training and validation (if ). See [callbacks](https://tensorflow.google.cn/api_docs/python/tf/keras/callbacks). Now available: `EarlyStopping` , `ModelCheckpoint`
        :return: A `History` object. Its `History.history` attribute is a record of training loss values and metrics values at successive epochs, as well as validation loss values and validation metrics values (if applicable).
        """
        if isinstance(x, dict):
            x = [x[feature] for feature in self.feature_index]

        do_validation = False
        if validation_data:
            do_validation = True
            if len(validation_data) == 2:
                val_x, val_y = validation_data
                val_sample_weight = None
            elif len(validation_data) == 3:
                val_x, val_y, val_sample_weight = validation_data  # pylint: disable=unpacking-non-sequence
            else:
                raise ValueError(
                    'When passing a `validation_data` argument, '
                    'it must contain either 2 items (x_val, y_val), '
                    'or 3 items (x_val, y_val, val_sample_weights), '
                    'or alternatively it could be a dataset or a '
                    'dataset or a dataset iterator. '
                    'However we received `validation_data=%s`' % validation_data)
            if isinstance(val_x, dict):
                val_x = [val_x[feature] for feature in self.feature_index]

        elif validation_split and 0. < validation_split < 1.:
            do_validation = True
            if hasattr(x[0], 'shape'):
                split_at = int(x[0].shape[0] * (1. - validation_split))
            else:
                split_at = int(len(x[0]) * (1. - validation_split))
            x, val_x = (slice_arrays(x, 0, split_at),
                        slice_arrays(x, split_at))
            y, val_y = (slice_arrays(y, 0, split_at),
                        slice_arrays(y, split_at))

        else:
            val_x = []
            val_y = []
        for i in range(len(x)):
            if len(x[i].shape) == 1:
                x[i] = np.expand_dims(x[i], axis=1)

        train_tensor_data = Data.TensorDataset(
            torch.from_numpy(
                np.concatenate(x, axis=-1)),
            torch.from_numpy(y))
        if batch_size is None:
            batch_size = 256

        model = self.train()
        loss_func = self.loss_func
        optim = self.optim

        if self.gpus:
            print('parallel running on these gpus:', self.gpus)
            model = torch.nn.DataParallel(model, device_ids=self.gpus)
            batch_size *= len(self.gpus)  # input `batch_size` is batch_size per gpu
        else:
            print(self.device)

        train_loader = DataLoader(
            dataset=train_tensor_data, shuffle=shuffle, batch_size=batch_size)

        sample_num = len(train_tensor_data)
        steps_per_epoch = (sample_num - 1) // batch_size + 1

        # configure callbacks
        callbacks = (callbacks or []) + [self.history]  # add history callback
        callbacks = CallbackList(callbacks)
        callbacks.set_model(self)
        callbacks.on_train_begin()
        callbacks.set_model(self)
        if not hasattr(callbacks, 'model'):  # for tf1.4
            callbacks.__setattr__('model', self)
        callbacks.model.stop_training = False

        # Train
        print("Train on {0} samples, validate on {1} samples, {2} steps per epoch".format(
            len(train_tensor_data), len(val_y), steps_per_epoch))
        for epoch in range(initial_epoch, epochs):
            callbacks.on_epoch_begin(epoch)
            epoch_logs = {}
            start_time = time.time()
            loss_epoch = 0
            total_loss_epoch = 0
            train_result = {}
            try:
                with tqdm(enumerate(train_loader), disable=verbose != 1) as t:
                    for _, (x_train, y_train) in t:
                        x = x_train.to(self.device).float()
                        if self.loss_func == F.cross_entropy:
                            y = y_train.to(self.device).long()
                        else:
                            y = y_train.to(self.device).float()

                        y_pred = model(x).squeeze()
                        y_pred = torch.sigmoid(y_pred)
                        optim.zero_grad()
                        
                        loss = loss_func(y_pred, y.squeeze(), reduction='sum')
                        reg_loss = self.get_regularization_loss()

                        total_loss = loss + reg_loss + self.aux_loss

                        loss_epoch += loss.item()
                        total_loss_epoch += total_loss.item()
                        total_loss.backward()
                        optim.step()

                        if verbose > 0:
                            for name, metric_fun in self.metrics.items():
                                if name not in train_result:
                                    train_result[name] = []
                                train_result[name].append(metric_fun(
                                    y.cpu().data.numpy(), y_pred.cpu().data.numpy().astype("float64")))


            except KeyboardInterrupt:
                t.close()
                raise
            t.close()

            # Add epoch_logs
            epoch_logs["loss"] = total_loss_epoch / sample_num
            for name, result in train_result.items():
                epoch_logs[name] = np.sum(result) / steps_per_epoch

            if do_validation:
                eval_result = self.evaluate(val_x, val_y, batch_size)
                for name, result in eval_result.items():
                    epoch_logs["val_" + name] = result
            # verbose
            if verbose > 0:
                epoch_time = int(time.time() - start_time)
                print('Epoch {0}/{1}'.format(epoch + 1, epochs))

                eval_str = "{0}s - loss: {1: .4f}".format(
                    epoch_time, epoch_logs["loss"])

                for name in self.metrics:
                    eval_str += " - " + name + \
                                ": {0: .4f}".format(epoch_logs[name])

                if do_validation:
                    for name in self.metrics:
                        eval_str += " - " + "val_" + name + \
                                    ": {0: .4f}".format(epoch_logs["val_" + name])
                print(eval_str)
            callbacks.on_epoch_end(epoch, epoch_logs)
            if self.stop_training:
                break

        callbacks.on_train_end()

        return self.history

    def evaluate(self, x, y, batch_size=256):
        """
        :param x: Numpy array of test data (if the model has a single input), or list of Numpy arrays (if the model has multiple inputs).
        :param y: Numpy array of target (label) data (if the model has a single output), or list of Numpy arrays (if the model has multiple outputs).
        :param batch_size: Integer or `None`. Number of samples per evaluation step. If unspecified, `batch_size` will default to 256.
        :return: Dict contains metric names and metric values.
        """
        pred_ans = self.predict(x, batch_size)
        eval_result = {}
        for name, metric_fun in self.metrics.items():
            eval_result[name] = metric_fun(y, pred_ans)
        return eval_result

    def predict(self, x, batch_size=256):
        """
        :param x: The input data, as a Numpy array (or list of Numpy arrays if the model has multiple inputs).
        :param batch_size: Integer. If unspecified, it will default to 256.
        :return: Numpy array(s) of predictions.
        """
        model = self.eval()
        if isinstance(x, dict):
            x = [x[feature] for feature in self.feature_index]
        for i in range(len(x)):
            if len(x[i].shape) == 1:
                x[i] = np.expand_dims(x[i], axis=1)

        tensor_data = Data.TensorDataset(
            torch.from_numpy(np.concatenate(x, axis=-1)))
        test_loader = DataLoader(
            dataset=tensor_data, shuffle=False, batch_size=batch_size)

        pred_ans = []
        with torch.no_grad():
            for _, x_test in enumerate(test_loader):
                x = x_test[0].to(self.device).float()

                y_pred = model(x).cpu().data.numpy()  # .squeeze()
                pred_ans.append(y_pred)

        return np.concatenate(pred_ans).astype("float64")

    def input_from_feature_columns(self, X, feature_columns, embedding_dict, support_dense=True):
        
        sparse_feature_columns = list(
            filter(lambda x: isinstance(x, SparseFeat), feature_columns)) if len(feature_columns) else []
        dense_feature_columns = list(
            filter(lambda x: isinstance(x, DenseFeat), feature_columns)) if len(feature_columns) else []

        varlen_sparse_feature_columns = list(
            filter(lambda x: isinstance(x, VarLenSparseFeat), feature_columns)) if feature_columns else []

        if not support_dense and len(dense_feature_columns) > 0:
            raise ValueError(
                "DenseFeat is not supported in dnn_feature_columns")

        sparse_embedding_list = [embedding_dict[feat.embedding_name](
            X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]].long()) for
            feat in sparse_feature_columns]
        sequence_embed_dict = varlen_embedding_lookup(X, self.embedding_dict, self.feature_index,
                                                      varlen_sparse_feature_columns)
        varlen_sparse_embedding_list = get_varlen_pooling_list(sequence_embed_dict, X, self.feature_index,
                                                               varlen_sparse_feature_columns, self.device)
        dense_value_list = [X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]] for feat in
                            dense_feature_columns]

        return sparse_embedding_list + varlen_sparse_embedding_list, dense_value_list

    def compute_input_dim(self, feature_columns, include_sparse=True, include_dense=True, feature_group=False):
        sparse_feature_columns = list(
            filter(lambda x: isinstance(x, (SparseFeat, VarLenSparseFeat)), feature_columns)) if len(
            feature_columns) else []
        dense_feature_columns = list(
            filter(lambda x: isinstance(x, DenseFeat), feature_columns)) if len(feature_columns) else []

        dense_input_dim = sum(
            map(lambda x: x.dimension, dense_feature_columns))
        if feature_group:
            sparse_input_dim = len(sparse_feature_columns)
        else:
            sparse_input_dim = sum(feat.embedding_dim for feat in sparse_feature_columns)
        input_dim = 0
        if include_sparse:
            input_dim += sparse_input_dim
        if include_dense:
            input_dim += dense_input_dim
        return input_dim

    def add_regularization_weight(self, weight_list, l1=0.0, l2=0.0):
        # For a Parameter, put it in a list to keep Compatible with get_regularization_loss()
        if isinstance(weight_list, torch.nn.parameter.Parameter):
            weight_list = [weight_list]
        # For generators, filters and ParameterLists, convert them to a list of tensors to avoid bugs.
        # e.g., we can't pickle generator objects when we save the model.
        else:
            weight_list = list(weight_list)
        self.regularization_weight.append((weight_list, l1, l2))

    def get_regularization_loss(self, ):
        total_reg_loss = torch.zeros((1,), device=self.device)
        for weight_list, l1, l2 in self.regularization_weight:
            for w in weight_list:
                if isinstance(w, tuple):
                    parameter = w[1]  # named_parameters
                else:
                    parameter = w
                if l1 > 0:
                    total_reg_loss += torch.sum(l1 * torch.abs(parameter))
                if l2 > 0:
                    try:
                        total_reg_loss += torch.sum(l2 * torch.square(parameter))
                    except AttributeError:
                        total_reg_loss += torch.sum(l2 * parameter * parameter)

        return total_reg_loss

    def add_auxiliary_loss(self, aux_loss, alpha):
        self.aux_loss = aux_loss * alpha

    def compile(self, optimizer,
                loss=None,
                metrics=None,
                ):
        """
        :param optimizer: String (name of optimizer) or optimizer instance. See [optimizers](https://pytorch.org/docs/stable/optim.html).
        :param loss: String (name of objective function) or objective function. See [losses](https://pytorch.org/docs/stable/nn.functional.html#loss-functions).
        :param metrics: List of metrics to be evaluated by the model during training and testing. Typically you will use `metrics=['accuracy']`.
        """
        self.metrics_names = ["loss"]
        self.optim = self._get_optim(optimizer)
        self.loss_func = self._get_loss_func(loss)
        self.metrics = self._get_metrics(metrics)

    def _get_optim(self, optimizer):
        if isinstance(optimizer, str):
            if optimizer == "sgd":
                optim = torch.optim.SGD(self.parameters(), lr=0.01)
            elif optimizer == "adam":
                optim = torch.optim.Adam(self.parameters())  # 0.001
            elif optimizer == "adagrad":
                optim = torch.optim.Adagrad(self.parameters())  # 0.01
            elif optimizer == "rmsprop":
                optim = torch.optim.RMSprop(self.parameters())
            else:
                raise NotImplementedError
        else:
            optim = optimizer
        return optim

    def _get_loss_func(self, loss):
        if isinstance(loss, str):
            if loss == "binary_crossentropy":
                loss_func = F.binary_cross_entropy
            elif loss == "mse":
                loss_func = F.mse_loss
            elif loss == "mae":
                loss_func = F.l1_loss
            elif loss == "cross_entropy":
                loss_func = F.cross_entropy
            else:
                raise NotImplementedError
        else:
            loss_func = loss
        return loss_func

    def _log_loss(self, y_true, y_pred, eps=1e-7, normalize=True, sample_weight=None, labels=None):
        # change eps to improve calculation accuracy
        return log_loss(y_true,
                        y_pred,
                        eps,
                        normalize,
                        sample_weight,
                        labels)

    def _get_metrics(self, metrics, set_eps=False):
        metrics_ = {}
        if metrics:
            for metric in metrics:
                if metric == "binary_crossentropy" or metric == "logloss":
                    if set_eps:
                        metrics_[metric] = self._log_loss
                    else:
                        metrics_[metric] = log_loss
                if metric == "auc":
                    metrics_[metric] = roc_auc_score
                if metric == "mse":
                    metrics_[metric] = mean_squared_error
                if metric == "accuracy" or metric == "acc":
                    metrics_[metric] = lambda y_true, y_pred: accuracy_score(
                        y_true, np.where(y_pred > 0.5, 1, 0))
                self.metrics_names.append(metric)
        return metrics_

    def _in_multi_worker_mode(self):
        # used for EarlyStopping in tf1.15
        return None

    @property
    def embedding_size(self, ):
        feature_columns = self.dnn_feature_columns
        sparse_feature_columns = list(
            filter(lambda x: isinstance(x, (SparseFeat, VarLenSparseFeat)), feature_columns)) if len(
            feature_columns) else []
        embedding_size_set = set([feat.embedding_dim for feat in sparse_feature_columns])
        if len(embedding_size_set) > 1:
            raise ValueError("embedding_dim of SparseFeat and VarlenSparseFeat must be same in this model!")
        return list(embedding_size_set)[0]




import torch


def concat_fun(inputs, axis=-1):
    def concat_fun(inputs, axis=-1):
        if len(inputs) == 1:
            return inputs[0]
        else:
            return torch.cat(inputs, dim=axis)

def combined_dnn_input(sparse_embedding_list, dense_value_list=[]):
    if len(sparse_embedding_list) > 0 and len(dense_value_list) > 0:
        sparse_dnn_input = torch.flatten(
            torch.cat(sparse_embedding_list, dim=-1), start_dim=1)
        dense_dnn_input = torch.flatten(
            torch.cat(dense_value_list, dim=-1), start_dim=1)
        return concat_fun([sparse_dnn_input, dense_dnn_input])
    elif len(sparse_embedding_list) > 0:
        return torch.flatten(torch.cat(sparse_embedding_list, dim=-1), start_dim=1)
    elif len(dense_value_list) > 0:
        return torch.flatten(torch.cat(dense_value_list, dim=-1), start_dim=1)
    else:
        raise NotImplementedError



import torch
import torch.nn as nn

# FM 二阶交互
class FactorizationMachines(nn.Module):
    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, X):
        # shape is: (batch_size, num_of_field, embedding_size)
        square_of_sum = torch.pow(
            torch.sum(X, dim=1,),
            2
        )
        sum_of_square = torch.sum(
            torch.pow(
                X, 2
            ),
            dim=1,
        ) # shape is (batch, embeding_size)
        # return shape is (batch, 1)
        return 0.5 * torch.sum(square_of_sum - sum_of_square, dim=1, keepdim=True)



import setuptools
import deepmatch_torch


with open("README.md", "r") as fh:
    long_description = fh.read()

# 去除了 PyTorch 的依赖，需要用户自己自己下载
REQUIRED_PACKAGES = [
    'tqdm', 'sklearn'
]

setuptools.setup(
    name="deepmatch-torch",
    version=deepmatch_torch.__version__,
    author="bbruceyuan",
    author_email="bruceyuan123@gmail.com",
    description="DeepMatch-Torch is a PyTorch Version deep matching model library for recommendations & advertising. It's easy to train models and to export representation vectors for user and item which can be used for ANN search.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bbruceyuan/DeepMatch-Torch",
    download_url='https://github.com/bbruceyuan/DeepMatch-Torch',
    packages=setuptools.find_packages(
        exclude=["tests", "tests.models", "tests.layers"]),
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*",  # '>=3.4',  # 3.4.6
    install_requires=REQUIRED_PACKAGES,
    extras_require={

    },
    entry_points={
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    license="MIT License",
    keywords=['match', "deepmatch_torch", "deepmatch-torch", "deepmatch-pytorch",
              'deep learning', 'torch', 'tensor', 'pytorch', 'deepmatch'],
)


from deepctr_torch.inputs import create_embedding_matrix

from deepctr_torch.inputs  import SparseFeat, DenseFeat, build_input_features
from deepctr_torch.layers.core import DNN

from .PLBaseModel import PLBaseModel
from ..utils import combined_dnn_input


class NCF(PLBaseModel):
    def __init__(self, user_feature_columns, item_feature_columns, 
                user_gmf_embedding_dim=20, item_gmf_embedding_dim=20,
                user_mlp_embedding_dim=20, item_mlp_embedding_dim=20, dnn_use_bn=False,
                dnn_hidden_units=[64, 32], dnn_activation='relu', l2_reg_dnn=0,
                dnn_dropout=0,
                l2_reg_linear=0.00001, l2_reg_embedding=0.00001, init_std=0.0001, seed=1024, task='binary', device='cpu', **kwargs):
        super().__init__(user_feature_columns, item_feature_columns, l2_reg_linear=l2_reg_linear, l2_reg_embedding=l2_reg_embedding, init_std=init_std, seed=seed, task=task, device=device, **kwargs)
        
        if len(user_feature_columns) > 1 or len(item_feature_columns) > 1:
            raise ValueError("目前 NCF 只支持 UserId 和 ItemId 作为特征")
        # 这里的 user_feature_columns 和 dssm 之类的不同，因为 gmf 和 mlp 部分的 embeding 层不一样
        self.dnn_hidden_units = dnn_hidden_units
        self.dnn_activation = dnn_activation
        self.l2_reg_dnn = l2_reg_dnn
        self.dnn_dropout = dnn_dropout
        self.user_mlp_embedding_dim= user_mlp_embedding_dim
        self.item_mlp_embedding_dim = item_mlp_embedding_dim
        self.dnn_use_bn = dnn_use_bn
        self.init_std = init_std
        self.l2_reg_embedding = l2_reg_embedding
        self.l2_reg_linear = l2_reg_linear
        self.seed = seed
        self.task = task 
        # self.device = device

        # 计算模型的复杂度
        self.user_gmf_embedding_dim = user_gmf_embedding_dim
        self.item_gmf_embedding_dim = item_gmf_embedding_dim

        # GMF Part
        self.user_gmf_feature_columns = [SparseFeat(feat, vocabulary_size=size, 
                                    embedding_dim=self.user_gmf_embedding_dim)
                                    for feat, size, *_ in self.user_feature_columns]
        self.item_gmf_feature_columns = [SparseFeat(feat, vocabulary_size=size, embedding_dim=self.item_gmf_embedding_dim)
                                    for feat, size, *_ in self.item_feature_columns]

        self.feature_index = build_input_features(self.user_gmf_feature_columns + self.item_gmf_feature_columns)

        self.gmf_embedding_dict = create_embedding_matrix(self.user_gmf_feature_columns + self.item_gmf_feature_columns, 
                                                        self.init_std, linear=False, sparse=False,
                                                        device=self.device)

        self.user_mlp_feature_columns = [SparseFeat(feat, vocabulary_size=size, embedding_dim=self.user_mlp_embedding_dim)
                                for feat, size, *_ in self.user_feature_columns]
        self.item_mlp_feature_columns = [SparseFeat(feat, vocabulary_size=size, embedding_dim=self.item_mlp_embedding_dim)
                                    for feat, size, *_ in self.item_feature_columns]
        
        self.mlp_embedding_dict = create_embedding_matrix(self.user_mlp_feature_columns + self.item_mlp_feature_columns, 
                                                        self.init_std, linear=False, sparse=False,
                                                        device=self.device)

        self.mlp_dnn = DNN(self.compute_input_dim(self.user_mlp_feature_columns + self.item_mlp_feature_columns), self.dnn_hidden_units, self.dnn_activation,
                            seed = self.seed, device=device)
        
        self.predict_layer = DNN(self.dnn_hidden_units[-1] + item_gmf_embedding_dim, hidden_units=[1], activation='sigmoid',seed=self.seed, device=device)

    def forward(self, X):
        user_gmf_sparse_embedding_list, user_gmf_dense_value_list = self.input_from_feature_columns(X,
                                                                                            self.user_gmf_feature_columns,
                                                                                            self.gmf_embedding_dict
                                                                                            )
        item_gmf_sparse_embedding_list, item_gmf_dense_value_list = self.input_from_feature_columns(X,
                                                                                            self.item_gmf_feature_columns,
                                                                                            self.gmf_embedding_dict)
        user_gmf_input = combined_dnn_input(user_gmf_sparse_embedding_list, user_gmf_dense_value_list)
        item_gmf_input = combined_dnn_input(item_gmf_sparse_embedding_list, item_gmf_dense_value_list)
        
        # 对应元素相差
        gmf_out = user_gmf_input * item_gmf_input
        
        # MLP part         
        user_mlp_sparse_embedding_list, user_mlp_dense_value_list = self.input_from_feature_columns(X,
                                                                                            self.user_mlp_feature_columns,
                                                                                            self.mlp_embedding_dict)
        item_mlp_sparse_embedding_list, item_mlp_dense_value_list = self.input_from_feature_columns(X,
                                                                                            self.item_mlp_feature_columns,
                                                                                            self.mlp_embedding_dict)
        user_mlp_input = combined_dnn_input(
            user_mlp_sparse_embedding_list, user_mlp_dense_value_list)
        item_mlp_input = combined_dnn_input(
            item_mlp_sparse_embedding_list, item_mlp_dense_value_list)

        mlp_input = combined_dnn_input([user_mlp_input, item_mlp_input])
        mlp_out = self.mlp_dnn(mlp_input)

        # Fusion of GMF and MLP
        neumf_input = combined_dnn_input([gmf_out, mlp_out])

        neumf_out = self.predict_layer(neumf_input)
        neumf_out = neumf_out.squeeze()
        return neumf_out



from .PLBaseModel import PLBaseModel
from deepctr_torch.layers import DNN
from deepctr_torch.inputs import SparseFeat, DenseFeat, VarLenSparseFeat
from deepctr_torch.inputs import get_varlen_pooling_list, varlen_embedding_lookup
from ..utils import combined_dnn_input
from ..inputs import create_embedding_matrix
import torch.nn.functional as F
from ..layers.core import CapsuleLayer
import torch


class MIND(PLBaseModel):
    def __init__(self, user_feature_columns, item_feature_columns, 
                num_sampled=5, user_dnn_hidden_units=[64, 32], 
                dnn_activation='relu', dnn_use_bn=False,
                dynamic_k=False, p=1, k_max=2,
                device="cpu", init_std=0.002,
               l2_reg_dnn=0, l2_reg_embedding=1e-6, 
               dnn_dropout=0, activation='relu', seed=1024, **kwargs):
        super(MIND, self).__init__(user_feature_columns, item_feature_columns, 
                l2_reg_linear=1e-5, l2_reg_embedding=1e-5,
                init_std=0.0001, seed=1024, task='binary', device='cpu', **kwargs)
        self.num_sampled = num_sampled
        self.pow_p = 2
        self.capsule_layer = CapsuleLayer(user_feature_columns[0].embedding_dim, 1, 
                    4, user_feature_columns[0].embedding_dim, 20)
        # self.label_att = LabelAwareAttention(4, 1)
        self.user_dnn = DNN(self.compute_input_dim(user_feature_columns), user_dnn_hidden_units,
                        activation=dnn_activation, init_std=init_std, device=device)

    def forward(self, X):
        
        # user_embedding = user_embedding.unsqueeze(1) # (batch, 1, embedding_dim)

        # item_embedding_list, _ = self.input_from_item_feature_columns(X, self.item_feature_columns, self.embedding_dict)
        # item_embedding = item_embedding_list[0]  # (batch, movie_list_len, feat_dim)
        user_embedding = self.user_tower(X)
        item_embedding = self.item_tower(X)

        if self.mode == "user_representation":
            return user_embedding
        if self.mode == "item_representation":
            return item_embedding
        # item_embedding_list 目前 size = 1
        # random sample 的数量
        batch_size = X.size(0)
        movie_list_len = item_embedding.shape[1]

        final_output_list = []
        for i in range(movie_list_len):
            # label attention_part
            tmp_input = [user_embedding, item_embedding[:, i, :]]
            keys = tmp_input[0]
            query = tmp_input[1].unsqueeze(1)

            weights = torch.sum(keys * query, axis=-1, keepdim=True)
            weights = torch.pow(weights, self.pow_p)
            weights = torch.softmax(weights, dim=1)
            output = torch.sum(keys * weights, axis=1)
            final_output_list.append(output)
        user_embedding = torch.cat(final_output_list, dim=1).view(batch_size, movie_list_len, -1)
        score = F.cosine_similarity(user_embedding, item_embedding, dim=-1)
        score = score.view(batch_size, -1)
        
        return score
    

    def item_tower(self, X):
        if self.mode == "user_representation":
            return None
        
        item_embedding_list, _ = self.input_from_item_feature_columns(X, self.item_feature_columns, self.embedding_dict)
        item_embedding = item_embedding_list[0]  # (batch, movie_list_len, feat_dim)
        return item_embedding

    def user_tower(self, X):
        if self.mode == "item_representation":
            return None
        # 找到 user_profile 相关的 feature
        # sample softmax 可以通过 构造样本实现
        user_sparse_feature_columns = [feat for feat in self.user_feature_columns if not feat.name.startswith("hist")]
        user_sparse_embedding_list, user_dense_value_list = \
            self.input_from_feature_columns(X, user_sparse_feature_columns, self.embedding_dict)
        
        # user history 序列特征
        user_history_feature_columns = [feat for feat in self.user_feature_columns if not feat.name.startswith("hist")]
        histroy_feature_embedding_list, _ = self.input_from_feature_columns(X, user_history_feature_columns, self.embedding_dict)

        capsule_input = histroy_feature_embedding_list[0]  # (batch, feat_embedding_dim)
        capsule_output = self.capsule_layer(capsule_input)
        
        cap_cnt = capsule_output.size(1) 
        user_dnn_input = combined_dnn_input(user_sparse_embedding_list, user_dense_value_list)

        user_dnn_input = torch.tile(user_dnn_input.unsqueeze(1), [1, cap_cnt, 1])
        user_dnn_input = torch.cat([user_dnn_input, capsule_output], dim=-1)

        user_embedding = self.user_dnn(user_dnn_input)  # (batch_size, interest_cnt, embedding_dim)
        return user_embedding

    def input_from_item_feature_columns(self, X, feature_columns, embedding_dict, support_dense=True):
    
        sparse_feature_columns = list(
            filter(lambda x: isinstance(x, SparseFeat), feature_columns)) if len(feature_columns) else []
        dense_feature_columns = list(
            filter(lambda x: isinstance(x, DenseFeat), feature_columns)) if len(feature_columns) else []

        varlen_sparse_feature_columns = list(
            filter(lambda x: isinstance(x, VarLenSparseFeat), feature_columns)) if feature_columns else []

        if not support_dense and len(dense_feature_columns) > 0:
            raise ValueError(
                "DenseFeat is not supported in dnn_feature_columns")

        sparse_embedding_list = [embedding_dict[feat.embedding_name](
            X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]].long()) for
            feat in sparse_feature_columns]

        # 这里返回的就是 movie_id 的 embedding
        sequence_embed_dict = varlen_embedding_lookup(X, self.embedding_dict, self.feature_index,
                                                      varlen_sparse_feature_columns)
        feat_name = varlen_sparse_feature_columns[0].name
        item_embedding = sequence_embed_dict[feat_name] 
        # shape is (batch, movie_id_len, feat_dim)

        varlen_sparse_embedding_list = [item_embedding]
        dense_value_list = [X[:, self.feature_index[feat.name][0]:self.feature_index[feat.name][1]] for feat in
                            dense_feature_columns]

        return sparse_embedding_list + varlen_sparse_embedding_list, dense_value_list